(* modify nothing before line 153 *)
open Ast

exception TypeError
exception RunTimeError
exception Unimplemented (* shows you what to change *)

(*
(* commented out, but an approach equivalent to the one below using
   explicit mutation
*)
(*********** environment implementation where env = var->exp **********)

(* Note: types do not matter at run-time, so we just use Ignore to fill 
   in for a type we do not care about *) 

let mt_env x = raise RunTimeError

let extend_env env x e = fun y -> if x=y then e else env y

(* call this from the Rec case of interpret *)
let extend_env_rec env f x e = 
  let rec new_env y =
    if y=f
    then Closure(Fun(x,Ignore,e), new_env)
    else env y
  in new_env

let lookup_env env y = env y
*)

(******* environment implementation where env = (var * (exp ref)) list *******)

(* Note: types do not matter at run-time, so we just use Ignore to fill 
   in for a type we do not care about *) 

let mt_env = []

let extend_env env x e = (x,ref e)::env

(* call this from the Rec case of interpret *)
let extend_env_rec env f x e =
  let r       = ref True in
  let new_env = (f,r)::env in
  let _       = r := Closure(Fun(x,Ignore,e),new_env) in
  new_env

let rec lookup_env env y =
  match env with
    [] -> raise RunTimeError
  | (x,r)::tl -> if x=y then !r else lookup_env tl y

(* type-checker -- we have done a couple cases for you *)
(* Note: t1=t2 will do what you want to compare types structurally, 
   e.g., Arrow(Int,Bool)=Arrow(Int,Bool)
   But t1==t2 is roughly pointer-equality -- do not accidentally use it or
   you will be very sorry *)

let mt_ctxt = []
let extend_ctxt ctxt x t = (x,t)::ctxt
let rec lookup_ctxt ctxt y =
  match ctxt with
    [] -> raise TypeError
  | (x,t)::tl -> if x=y then t else lookup_ctxt tl y

let rec typecheck ctxt e =
  match e with
    True  -> Bool
  | False -> Bool
  | If(e1,e2,e3) ->
      let t1 = typecheck ctxt e1 in
      let t2 = typecheck ctxt e2 in
      let t3 = typecheck ctxt e3 in
      if (t1=Bool) && (t2=t3)
      then t2
      else raise TypeError
  | Const i -> Int
  | Plus(e1,e2) ->
      let t1 = typecheck ctxt e1 in
      let t2 = typecheck ctxt e2 in
      if (t1=Int) && (t2=Int)
      then Int
      else raise TypeError
  | GreaterThan(e1,e2) ->
      let t1 = typecheck ctxt e1 in
      let t2 = typecheck ctxt e2 in
      if (t1=Int) && (t2=Int)
      then Bool
      else raise TypeError
  | Var(x) -> lookup_ctxt ctxt x
  | Fun(x,t,e) ->
      let t1 = typecheck (extend_ctxt ctxt x t) e in
      Arrow(t,t1)
  | Apply(e1,e2) ->
      let t1 = typecheck ctxt e1 in
      let t2 = typecheck ctxt e2 in
      (match t1 with
	Arrow(t3,t4) -> if t3=t2 then t4 else raise TypeError
      |	_ -> raise TypeError)
  | Rec(f,t,x,e) ->
      (match t with
	Arrow(t1,t2) -> 
	  let c1 = extend_ctxt ctxt f t  in
	  let c2 = extend_ctxt c1 x t1 in
	  let t3 = typecheck c2 e in
	  if t3=t2
	  then t
	  else raise TypeError
      |	_ -> raise TypeError)
  | Closure(_,_) -> raise TypeError

(* evaluator 
   remember we are large-step -- return a value
   *)
let rec interpret env e =
  match e with
    True -> e
  | False -> e
  | If(e1,e2,e3) ->
      let v1 = interpret env e1 in
      (match v1 with
	True -> interpret env e2
      |	False -> interpret env e3
      |	_ -> raise RunTimeError)
  | Const i -> e
  | Closure(_,_) -> e
  | Plus(e1,e2) ->
      let v1 = interpret env e1 in
      let v2 = interpret env e2 in
      (match (v1,v2) with
	(Const i1, Const i2) -> Const(i1+i2)
      |	_ -> raise RunTimeError)
  | GreaterThan(e1,e2) ->
      let v1 = interpret env e1 in
      let v2 = interpret env e2 in
      (match (v1,v2) with
	(Const i1, Const i2) -> if i1 > i2 then True else False
      |	_ -> raise RunTimeError)
  | Var(x) -> lookup_env env x
  | Fun(_,_,_) -> Closure(e,env)
  | Apply(e1,e2) ->
      let v1 = interpret env e1 in
      let v2 = interpret env e2 in
      (match v1 with
	Closure(Fun(x,_,e3),envc) -> interpret (extend_env envc x v2) e3
      |	_ -> raise RunTimeError)
  | Rec(f,t,x,e) ->
      Closure(Fun(x,Ignore,e),extend_env_rec env f x e)

type tcexp = exp (* clients do not know this! *)

(**** Modify nothing above this line ****)
let interpret1 e = raise Unimplemented
let typecheck2 e = raise Unimplemented
let interpret2 e = raise Unimplemented
let typecheck3 e = raise Unimplemented
let typecheck4 e = raise Unimplemented
let interpret4 e = raise Unimplemented

